# all the objects used with the db session

class SelectObject:
    def __init__(self, query, params=None, fetch_mode='all'):
        self.type = 'SELECT'
        self.query = query
        self.params = params
        self.fetch_mode = fetch_mode

    def execute(self, cursor):
        if self.params:
            cursor.execute(self.query, tuple(self.params))
        else:
            cursor.execute(self.query)

        if self.fetch_mode == 'one':
            data = cursor.fetchone()
        else:
            data = cursor.fetchall()
        return data


# used for example where or querys to prevent the same key word error if you want to do something like
# username='test1' or username='test2' you would do
# username1=FilterObject('username', 'test1') or username2=FilterObject('username', 'test2')
class FilterObject:
    def __init__(self, **args):
        self.args_to_column_and_value(args)

    def args_to_column_and_value(self, args):
        for arg in args:
            self.column = arg
            self.value = args[arg]


class InsertObject:
    def __init__(self, query, params=None, commit=False):
        self.type = 'INSERT'
        self.query = query
        self.params = params
        self.commit = commit

    def execute(self, conn, cursor):
        if self.params:
            cursor.execute(self.query, tuple(self.params))
        else:
            cursor.execute(self.query)
        if self.commit:
            conn.commit()


class UpdateObject:
    def __init__(self, query, params=None, commit=False):
        self.type = 'UPDATE'
        self.query = query
        self.params = params
        self.commit = commit

    def execute(self, conn, cursor):
        if self.params:
            cursor.execute(self.query, tuple(self.params))
        else:
            cursor.execute(self.query)
        if self.commit:
            conn.commit()


class DeleteObject:
    def __init__(self, query, params=None, commit=False):
        self.type = 'DELETE'
        self.query = query
        self.params = params
        self.commit = commit

    def execute(self, conn, cursor):
        if self.params:
            cursor.execute(self.query, tuple(self.params))
        else:
            cursor.execute(self.query)
        if self.commit:
            conn.commit()



from .make_migrations import MakeMigrationTests
from .migrate import MigrateTests
from .select import Select
from .insert import Insert
from .update import Update
from .delete import Delete
from flask import Flask
import os
import shutil
from .req import passed_str


def reset(app):
    if os.path.isdir('migrations'):
        shutil.rmtree('migrations')
    if os.path.isfile(app.config['DB_PATH']):
        os.remove(app.config['DB_PATH'])


def run():
    app = Flask(__name__)
    app.config['DB_PATH'] = 'db.db'

    # tests = [Delete(app)]
    tests = [MakeMigrationTests(app), MigrateTests(app), Select(app), Insert(app), Update(app), Delete(app)]
    output = ""
    total_tests_run = 0
    total_tests_passed = 0
    for test in tests:
        output += test.name + ':\n'
        test_output, tests_run, tests_passed = test.run()
        output += test_output
        tests_passed_fraction = passed_str(f"{tests_passed}/{tests_run}")
        tests_passed_percent = passed_str(f"{(round(tests_passed / tests_run, 4) if tests_passed > 0 else 0) * 100}%")
        output += f" - {test.name} Tests Passed {tests_passed_fraction} or {tests_passed_percent}\n"
        total_tests_run += tests_run
        total_tests_passed += tests_passed
        reset(app)
    os.system('cls')
    tests_passed_fraction = passed_str(f"{total_tests_passed}/{total_tests_run}")
    tests_passed_percent = passed_str(f"{(round(total_tests_passed / total_tests_run, 4) if total_tests_passed > 0 else 0) * 100}%")
    output += f"Tests Passed {tests_passed_fraction} or {tests_passed_percent}\n"

    return output



from .query_objects import SelectObject, FilterObject
import math


class ForeignRowObject:
    def __init__(self, session, foreign_model, tuple_rows, value):
        self.session = session
        self.tuple_rows = tuple_rows
        self.foreign_model = foreign_model
        self.columns_unsorted = self.foreign_model.columns
        self.value = value
        self.get_columns()
        self.rows_to_rowobjects()

    def rows_to_rowobjects(self):
        self.rows = []
        for row in self.tuple_rows:
            self.rows.append(RowObject(self.session, row, self.columns))

    def get_columns(self):
        table_data = self.session.select(selectobj=SelectObject(f"PRAGMA table_info({self.foreign_model.model_name});", fetch_mode='all'))
        self.columns = {}
        for column in table_data:
            for col in self.columns_unsorted:
                if self.columns_unsorted[col].name == column[1]:
                    self.columns[column[1]] = self.columns_unsorted[col]


class RowObject:
    def __init__(self, session, row_data, columns):
        self.session = session
        self.row_data = row_data
        self.columns = columns
        self.row_to_var()

    def row_to_var(self):
        for i in range(len(self.columns)):
            # if the column is a foreign key
            if self.columns[list(self.columns.keys())[i]].fk:
                foreign_model = self.columns[list(self.columns.keys())[i]].foreign_model_in
                foreign_pk = foreign_model.primary_key.name
                _locals = locals()
                # get the row that the foreign key links to
                foreign_data = self.session.select(SelectObject(f"SELECT * FROM {foreign_model.model_name} WHERE {foreign_pk} = ?", params=[self.row_data[i]], fetch_mode='all'))
                exec(f"self.{list(self.columns.keys())[i]} = value", {'self': self, 'value': ForeignRowObject(self.session, foreign_model, foreign_data, self.row_data[i])})
            # if the column is a boolean column change the value from a int to a bool
            elif self.columns[list(self.columns.keys())[i]].python_type == bool:
                exec(f"self.{list(self.columns.keys())[i]} = value", {'self': self, 'value': bool(self.row_data[i])})

            else:
                exec(f"self.{list(self.columns.keys())[i]} = value", {'self': self, 'value': self.row_data[i]})


class PaginationPage:
    def __init__(self, session, table_name, columns, query, params, per_page, page_num):
        self.session = session
        self.table_name = table_name
        self.columns = columns
        self.query = query
        self.params = params
        self.per_page = per_page
        self.page_num = page_num
        self.get_all_data()
        self.get_items()

    def get_all_data(self):
        # work out how many posible pages there are
        result = self.session.select(self.session, SelectObject(self.query, params=self.params, fetch_mode='all'))
        self.page_amount = math.ceil(len(result) / self.per_page)

    def get_items(self):
        # get the page data
        start = self.per_page * (self.page_num - 1)
        self.query = self.query + " LIMIT ? OFFSET ?"
        self.params.append(self.per_page)
        self.params.append(start)
        items = self.session.select(self.session, SelectObject(self.query, params=self.params, fetch_mode='all'))
        f_result = []
        if items != []:
            for row in items:
                f_result.append(RowObject(self.session, row, self.columns))
            self.items = f_result

    # button stuff
    def iter_page(self, first, last, left, right):
        self.page_nums = []
        self.page_nums.append(1)
        for i in range(left):
            if self.page_num - (i + 1) > 1:
                self.page_nums.append(self.page_num - (i + 1))
        if self.page_num != 1 and self.page_num != self.page_amount:
            self.page_nums.append(self.page_num)
        for i in range(right):
            if self.page_num + (i + 1) < self.page_amount:
                self.page_nums.append(self.page_num + (i + 1))
        if self.page_amount != 1:
            self.page_nums.append(self.page_amount)
        return self.page_nums


class FinalFunctions:
    def __init__(self, session, table_name, columns, query, params):
        self.session = session
        self.table_name = table_name
        self.columns = columns
        self.query = query
        self.params = params

    def all(self):
        result = self.session.select(SelectObject(self.query, params=self.params, fetch_mode='all'))
        if result:
            final_result = []
            for row in result:
                final_result.append(RowObject(self.session, row, self.columns))
            return final_result
        else:
            return []

    def one(self):
        result = self.session.select(SelectObject(self.query, params=self.params, fetch_mode='one'))
        if result:
            return RowObject(self.session, result, self.columns)
        else:
            # returning a RowObject with all values set to None
            none_result = []
            for column in self.columns:
                none_result.append(None)
            return RowObject(self.session, tuple(none_result), self.columns)

    def paginate(self, per_page, page_num):
        return PaginationPage(self.table_name, self.columns, self.query, self.params, per_page, page_num)


class SortBy(FinalFunctions):
    def sortby(self, **args):
        # used the function is being run from the sortby function in Select
        if 'pre_data' in list(args.keys()):
            args = args['pre_data']
        # adding to the query
        self.query += ' ORDER BY '
        for arg in args:
            self.query += f"{arg} {args[arg]}, "
        # taking off extra comma and space
        self.query = self.query[:-2]
        return FinalFunctions(self.session, self.table_name, self.columns, self.query, self.params)


class Filter(FinalFunctions):
    def filter(self, conjunctive_op='AND', comparison_op='=', **args):
        if 'pre_data' in list(args.keys()):
            # used the function is being run from the filter function in Select
            args = args['pre_data']
            # checking if a conjunctive_op was passed from the filter function in Select
            if 'conjunctive_op' in list(args.keys()):
                conjunctive_op = args['conjunctive_op']
                del args['conjunctive_op']
            if 'comparison_op' in list(args.keys()):
                comparison_op = args['comparison_op']
                del args['comparison_op']
        # adding to the query
        self.query += " WHERE "
        self.params = []
        for param in args:
            # if the param is a FilterObject get the column name and value from that obj
            if isinstance(args[param], FilterObject):
                self.query += f"{args[param].column} {comparison_op} ? {conjunctive_op} "
                self.params.append(args[param].value)
            else:
                self.query += f"{param} {comparison_op} ? {conjunctive_op} "
                self.params.append(args[param])
        # taking off the extra spaces and conjunctive_op at the end
        self.query = self.query[:-(len(conjunctive_op) + 2)]
        return SortBy(self.session, self.table_name, self.columns, self.query, self.params)


class Select(FinalFunctions):
    def __init__(self, session, table_name, columns):
        self.session = session
        self.table_name = table_name
        self.columns_unsorted = columns
        self.get_columns()
        self.query = f"SELECT * FROM {self.table_name}"
        self.params = []

    def get_columns(self):
        table_data = self.session.select(selectobj=SelectObject(f"PRAGMA table_info({self.table_name});", fetch_mode='all'))
        self.columns = {}
        for column in table_data:
            for col in self.columns_unsorted:
                if self.columns_unsorted[col].name == column[1]:
                    self.columns[column[1]] = self.columns_unsorted[col]

    def sortby(self, **kwargs):
        return SortBy(self.session, self.table_name, self.columns, self.query, self.params).sortby(pre_data=kwargs)

    def filter(self, **kwargs):
        return Filter(self.session, self.table_name, self.columns, self.query, self.params).filter(pre_data=kwargs)



from .query_objects import UpdateObject, FilterObject


class RunUpdate:
    def __init__(self, session, query, params, commit):
        self.session = session
        self.query = query
        self.params = params
        self.commit = commit

    def all(self):
        if self.commit:
            self.session.add(UpdateObject(self.query, params=self.params, commit=self.commit))
        else:
            return UpdateObject(self.query, params=self.params, commit=self.commit)

    def where(self, conjunctive_op='AND', comparison_op='=', **args):
        self.query += " WHERE "
        for param in args:
            # if the param is a FilterObject get the column name and value from that obj
            if isinstance(args[param], FilterObject):
                self.query += f"{args[param].column} {comparison_op} ? {conjunctive_op} "
                self.params.append(args[param].value)
            else:
                self.query += f"{param} {comparison_op} ? {conjunctive_op} "
                self.params.append(args[param])
        # taking off the extra spaces and conjunctive_op at the end
        self.query = self.query[:-(len(conjunctive_op) + 2)]
        if self.commit:
            self.session.add(UpdateObject(self.query, params=self.params, commit=self.commit))
        else:
            return UpdateObject(self.query, params=self.params, commit=self.commit)


class Update:
    def __init__(self, session, table_name):
        self.session = session
        self.table_name = table_name

    def set(self, commit=False, **args):
        set_str = ""
        params = []
        for arg in args:
            set_str += f"{arg} = ?, "
            params.append(args[arg])
        query = f"UPDATE {self.table_name} SET {set_str[:-2]}"
        return RunUpdate(self.session, query, params, commit)



from colorama import Fore, init, Back
init()
passed = Fore.GREEN + 'Passed' + Fore.RESET
failed = Fore.RED + 'Failed' + Fore.RESET


def get_default_value(column):
    column = column()
    if column.python_type == str:
        return 'test-string'
    if column.python_type == int:
        return 1234567890
    if column.python_type == float:
        return 100.001
    if column.python_type == bool:
        return False


def passed_str(str):
    return Fore.GREEN + str + Fore.RESET


def failed_str(str):
    return Fore.RED + str + Fore.RESET


def get_test_str(test_name, passed_test):
    return f" - {test_name} - {passed if passed_test else failed}\n"



from .query_objects import InsertObject


class Insert:
    def __init__(self, session, table_name):
        self.session = session
        self.table_name = table_name

    def add(self, commit=False, **args):
        # making the query
        columns = ""
        value_str = ""
        params = []
        for arg in args:
            columns += arg + ', '
            value_str += '?, '
            params.append(args[arg])
        query = f"INSERT INTO {self.table_name} ({columns[:-2]}) VALUES ({value_str[:-2]})"
        if commit:
            self.session.add(InsertObject(query, params=params, commit=commit))
        else:
            return InsertObject(query, params=params, commit=commit)



import sqlite3
from queue import Queue
import threading


# this is used so that in run_queue we can check the .type of the object passed
class CommitObject:
    def __init__(self):
        self.type = 'Commit'


class Session:
    def __init__(self, db_path):
        self.db_path = db_path

        # making db connection
        self.conn = sqlite3.connect(self.db_path, check_same_thread=False)
        self.c = self.conn.cursor()

        # making running thread
        self.queue = Queue(maxsize=0)
        self.done = {}
        self.running = True
        self.t = threading.Thread(target=self.run_queue)
        self.t.start()

    def run_queue(self):
        while self.running:
            item = self.queue.get()
            # checking if its a select query to return the result
            if item.type == 'SELECT':
                data = item.execute(self.c)
                self.done[str(item)] = data
            elif item.type == 'Commit':
                self.conn.commit()
                self.done[str(item)] = None
            else:
                item.execute(self.conn, self.c)
                self.done[str(item)] = None

    def select(self, selectobj):
        if self.running:
            self.queue.put(selectobj)
            # waiting for the query to be run then returning it
            while str(selectobj) not in list(self.done.keys()):
                pass
            data = self.done[str(selectobj)]
            del self.done[str(selectobj)]
            return data

    def add(self, obj):
        if self.running:
            # adding a insert, update or delete object to the queue for the run_queue thread to execute
            self.queue.put(obj)
            while str(obj) not in list(self.done.keys()):
                pass
            del self.done[str(obj)]

    def commit(self):
        if self.running:
            obj = CommitObject()
            self.queue.put(obj)
            while str(obj) not in list(self.done.keys()):
                pass
            del self.done[str(obj)]

    def stop(self):
        # stopping queue thread and closing db
        self.running = False
        self.conn.close()



from .. import DB, Model
from ..columns import (PrimaryKeyColumn, ForeignKeyColumn,
                       StringColumn, IntegerColumn,
                       FloatColumn, BooleanColumn)
from .req import get_test_str, get_default_value
import os
import shutil
import sqlite3


class UpdateAllTest:
    def __init__(self, app, commit):
        self.test_name = 'UpdateAllTest'
        self.app = app
        self.commit = commit

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        c.execute("INSERT INTO test1 (username, password) VALUES (?, ?)", ['username1', 'password1'])
        conn.commit()
        conn.close()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        if self.commit:
            Test1.update(username='username2', commit=True).all()
        else:
            obj = Test1.update(username='username2').all()
            self.db.session.add(obj)
            self.db.session.commit()
        self.db.session.stop()

        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("SELECT * FROM test1")
        data = c.fetchone()
        conn.close()
        if data[1] != 'username2':
            return False
        return True


class UpdateWhereTest:
    def __init__(self, app, commit):
        self.test_name = 'UpdateAllTest'
        self.app = app
        self.commit = commit

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        c.execute("INSERT INTO test1 (username, password) VALUES (?, ?)", ['username1', 'password1'])
        c.execute("INSERT INTO test1 (username, password) VALUES (?, ?)", ['username2', 'password2'])
        conn.commit()
        conn.close()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        if self.commit:
            Test1.update(username='username3', commit=True).where(password='password1')
        else:
            obj = Test1.update(username='username3').where(password='password1')
            self.db.session.add(obj)
            self.db.session.commit()
        self.db.session.stop()

        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("SELECT * FROM test1 WHERE username = ?", ('username3',))
        data = c.fetchall()
        conn.close()
        if len(data) != 1:
            return False
        return True


class Update:
    def __init__(self, app):
        self.name = 'UpdateTests'
        self.app = app
        self.tests = [UpdateAllTest(self.app, True),
                      UpdateAllTest(self.app, False),
                      UpdateWhereTest(self.app, True),
                      UpdateWhereTest(self.app, False)]

    def run(self):
        output = ""
        tests_run = 0
        tests_passed = 0
        for test in self.tests:
            tests_run += 1
            try:
                result = test.run()
                output += get_test_str(test.test_name, result)
                if result:
                    tests_passed += 1
            except Exception as e:
                test.run()
        return output, tests_run, tests_passed



from .. import DB, Model
from ..columns import (PrimaryKeyColumn, ForeignKeyColumn,
                       StringColumn, IntegerColumn,
                       FloatColumn, BooleanColumn)
from .req import get_test_str, get_default_value
import os
import shutil
import sqlite3


single_model_make_migrations = {
    StringColumn: '''
from sql_cable.migrations import Migration


class test(Migration):
    def __init__(self):
        self.table_name = 'test'
        self.table_creation_str = 'CREATE TABLE test ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [c] STRING, [cd] STRING DEFAULT [test-string], [cdl] STRING(3, 16) DEFAULT [test-string], [cl] STRING(3, 16), [cn] STRING NOT NULL, [cnd] STRING NOT NULL DEFAULT [test-string], [cnl] STRING(3, 16) NOT NULL, [cu] STRING UNIQUE, [cud] STRING UNIQUE DEFAULT [test-string], [cul] STRING(3, 16) UNIQUE, [cun] STRING UNIQUE NOT NULL, [cundl] STRING(3, 16) UNIQUE NOT NULL DEFAULT [test-string])'
        self.changes = ['test table does not exist']
        self.columns = ['c', 'cd', 'cdl', 'cl', 'cn', 'cnd', 'cnl', 'cu', 'cud', 'cul', 'cun', 'cundl', 'id']
        self.action = 'CREATE'


models = [test()]''',
    IntegerColumn: '''
from sql_cable.migrations import Migration


class test(Migration):
    def __init__(self):
        self.table_name = 'test'
        self.table_creation_str = 'CREATE TABLE test ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [c] INTEGER, [cd] INTEGER DEFAULT 1234567890, [cdl] INTEGER(3, 16) DEFAULT 1234567890, [cl] INTEGER(3, 16), [cn] INTEGER NOT NULL, [cnd] INTEGER NOT NULL DEFAULT 1234567890, [cnl] INTEGER(3, 16) NOT NULL, [cu] INTEGER UNIQUE, [cud] INTEGER UNIQUE DEFAULT 1234567890, [cul] INTEGER(3, 16) UNIQUE, [cun] INTEGER UNIQUE NOT NULL, [cundl] INTEGER(3, 16) UNIQUE NOT NULL DEFAULT 1234567890)'
        self.changes = ['test table does not exist']
        self.columns = ['c', 'cd', 'cdl', 'cl', 'cn', 'cnd', 'cnl', 'cu', 'cud', 'cul', 'cun', 'cundl', 'id']
        self.action = 'CREATE'


models = [test()]''',
    FloatColumn: '''
from sql_cable.migrations import Migration


class test(Migration):
    def __init__(self):
        self.table_name = 'test'
        self.table_creation_str = 'CREATE TABLE test ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [c] REAL, [cd] REAL DEFAULT 100.001, [cdl] REAL(3, 16) DEFAULT 100.001, [cl] REAL(3, 16), [cn] REAL NOT NULL, [cnd] REAL NOT NULL DEFAULT 100.001, [cnl] REAL(3, 16) NOT NULL, [cu] REAL UNIQUE, [cud] REAL UNIQUE DEFAULT 100.001, [cul] REAL(3, 16) UNIQUE, [cun] REAL UNIQUE NOT NULL, [cundl] REAL(3, 16) UNIQUE NOT NULL DEFAULT 100.001)'
        self.changes = ['test table does not exist']
        self.columns = ['c', 'cd', 'cdl', 'cl', 'cn', 'cnd', 'cnl', 'cu', 'cud', 'cul', 'cun', 'cundl', 'id']
        self.action = 'CREATE'


models = [test()]''',
    BooleanColumn: '''
from sql_cable.migrations import Migration


class test(Migration):
    def __init__(self):
        self.table_name = 'test'
        self.table_creation_str = 'CREATE TABLE test ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [c] BOOLEAN, [cd] BOOLEAN, [cdl] BOOLEAN(3, 16), [cl] BOOLEAN(3, 16), [cn] BOOLEAN NOT NULL, [cnd] BOOLEAN NOT NULL, [cnl] BOOLEAN(3, 16) NOT NULL, [cu] BOOLEAN UNIQUE, [cud] BOOLEAN UNIQUE, [cul] BOOLEAN(3, 16) UNIQUE, [cun] BOOLEAN UNIQUE NOT NULL, [cundl] BOOLEAN(3, 16) UNIQUE NOT NULL)'
        self.changes = ['test table does not exist']
        self.columns = ['c', 'cd', 'cdl', 'cl', 'cn', 'cnd', 'cnl', 'cu', 'cud', 'cul', 'cun', 'cundl', 'id']
        self.action = 'CREATE'


models = [test()]'''
}


class SingleModelMakeMigrations:
    def __init__(self, app, column_type):
        self.app = app
        self.column_type = column_type
        self.test_name = f'''SingleModelMakeMigrations_{str(self.column_type).split("<class 'sql_cable.columns.")[1].split("'>")[0]}'''

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        self.db = DB(self.app)

        class Test(Model):
            id = PrimaryKeyColumn()
            c = self.column_type()
            cu = self.column_type(unique=True)
            cn = self.column_type(not_null=True)
            cd = self.column_type(default=get_default_value(self.column_type))
            cl = self.column_type(min_length=3, max_length=16)

            cun = self.column_type(unique=True, not_null=True)
            cud = self.column_type(unique=True, default=get_default_value(self.column_type))
            cul = self.column_type(unique=True, min_length=3, max_length=16)
            cnd = self.column_type(not_null=True, default=get_default_value(self.column_type))
            cnl = self.column_type(not_null=True, min_length=3, max_length=16)
            cdl = self.column_type(default=get_default_value(self.column_type), min_length=3, max_length=16)

            cundl = self.column_type(unique=True, not_null=True,
                                     default=get_default_value(self.column_type),
                                     min_length=3, max_length=16)
        self.db.register_model(Test)
        path = self.db.make_migrations(test=True)
        self.db.session.stop()

        with open(path, 'r') as f:
            migrations_file_data = f.read()
        migrations_file_data = migrations_file_data.replace(migrations_file_data.split('\n')[0], '')
        if migrations_file_data == single_model_make_migrations[self.column_type]:
            return True
        # print(migrations_file_data)
        return False


double_model_make_migrations_correct = '''
from sql_cable.migrations import Migration


class test1(Migration):
    def __init__(self):
        self.table_name = 'test1'
        self.table_creation_str = 'CREATE TABLE test1 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [name] STRING)'
        self.changes = ['test1 table does not exist']
        self.columns = ['id', 'name']
        self.action = 'CREATE'


class test2(Migration):
    def __init__(self):
        self.table_name = 'test2'
        self.table_creation_str = 'CREATE TABLE test2 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [test_1_id] INTEGER REFERENCES test1 (id)  ON DELETE CASCADE)'
        self.changes = ['test2 table does not exist']
        self.columns = ['id', 'test_1_id']
        self.action = 'CREATE'


models = [test1(), test2()]'''


class DoubleModelMakeMigrations:
    def __init__(self, app):
        self.test_name = 'DoubleModelMakeMigrations'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            name = StringColumn()

        class Test2(Model):
            id = PrimaryKeyColumn()
            test_1_id = ForeignKeyColumn(Test1)

        self.db.register_model(Test1)
        self.db.register_model(Test2)

        path = self.db.make_migrations(test=True)
        self.db.session.stop()
        with open(path, 'r') as f:
            migrations_file_data = f.read()
        migrations_file_data = migrations_file_data.replace(migrations_file_data.split('\n')[0], '')
        if migrations_file_data == double_model_make_migrations_correct:
            return True
        return False


single_model_update_make_migrations_correct = '''
from sql_cable.migrations import Migration


class test1(Migration):
    def __init__(self):
        self.table_name = 'test1'
        self.table_creation_str = 'CREATE TABLE test1 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [cd] STRING DEFAULT [correct value], [name] STRING NOT NULL, [password] STRING(3, 16), [unique] STRING UNIQUE)'
        self.changes = ["cd's default value is meant to be [correct value] not notcorrect", 'name is meant have not_null', 'password is meant to have a type of STRING(3, 16) not STRING (1, 8)']
        self.columns = ['cd', 'id', 'name', 'password', 'unique']
        self.action = 'UPDATE'


models = [test1()]'''


class SingleModelUpdateMakeMigrations:
    def __init__(self, app):
        self.test_name = 'SingleModelUpdateMakeMigrations'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, name STRING, password STRING (1, 8), [unique] STRING UNIQUE, cd STRING DEFAULT notcorrect);")
        conn.commit()
        conn.close()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            name = StringColumn(not_null=True)
            password = StringColumn(min_length=3, max_length=16)
            unique = StringColumn(unique=True)
            cd = StringColumn(default='correct value')

        self.db.register_model(Test1)
        path = self.db.make_migrations(test=True)
        self.db.session.stop()
        with open(path, 'r') as f:
            migrations_file_data = f.read()
        migrations_file_data = migrations_file_data.replace(migrations_file_data.split('\n')[0], '')
        if migrations_file_data == single_model_update_make_migrations_correct:
            return True
        return False


two_model_update_make_migrations_correct = '''
from sql_cable.migrations import Migration


class test1(Migration):
    def __init__(self):
        self.table_name = 'test1'
        self.table_creation_str = 'CREATE TABLE test1 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [password] STRING, [username] STRING UNIQUE)'
        self.changes = ['password is meant to have a type of STRING not STRING (1, 8)', 'username does not exist']
        self.columns = ['id', 'password', 'username']
        self.action = 'UPDATE'


class test2(Migration):
    def __init__(self):
        self.table_name = 'test2'
        self.table_creation_str = 'CREATE TABLE test2 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [test1_id] INTEGER REFERENCES test1 (id)  ON DELETE CASCADE)'
        self.changes = ["test1_id's on_delete is meant to be CASCADE but is SET NULL", "test1_id's on_update is meant to be NONE but is SET DEFAULT"]
        self.columns = ['id', 'test1_id']
        self.action = 'UPDATE'


models = [test1(), test2()]'''


class TwoModelUpdateMakeMigrations:
    def __init__(self, app):
        self.test_name = 'TwoModelUpdateMakeMigrations'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, name STRING, password STRING (1, 8), [unique] STRING UNIQUE, cd STRING DEFAULT notcorrect);")
        c.execute("CREATE TABLE test2 (id INTEGER PRIMARY KEY AUTOINCREMENT, test1_id INTEGER REFERENCES test1 (id) ON DELETE SET NULL ON UPDATE SET DEFAULT);")
        conn.commit()
        conn.close()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn(unique=True)
            password = StringColumn()

        class Test2(Model):
            id = PrimaryKeyColumn()
            test1_id = ForeignKeyColumn(Test1)

        self.db.register_model(Test1)
        self.db.register_model(Test2)
        path = self.db.make_migrations(test=True)
        self.db.session.stop()
        with open(path, 'r') as f:
            migrations_file_data = f.read()
        migrations_file_data = migrations_file_data.replace(migrations_file_data.split('\n')[0], '')
        if migrations_file_data == two_model_update_make_migrations_correct:
            return True
        return False


class MakeMigrationTests:
    def __init__(self, app):
        self.name = 'MakeMigrationTests'
        self.app = app
        self.tests = [SingleModelMakeMigrations(self.app, StringColumn),
                      SingleModelMakeMigrations(self.app, IntegerColumn),
                      SingleModelMakeMigrations(self.app, FloatColumn),
                      SingleModelMakeMigrations(self.app, BooleanColumn),
                      DoubleModelMakeMigrations(self.app),
                      SingleModelUpdateMakeMigrations(self.app),
                      TwoModelUpdateMakeMigrations(self.app)]

    def run(self):
        output = ""
        tests_run = 0
        tests_passed = 0
        for test in self.tests:
            tests_run += 1
            try:
                result = test.run()
                output += get_test_str(test.test_name, result)
                if result:
                    tests_passed += 1
            except Exception as e:
                test.run()
        return output, tests_run, tests_passed



from .columns import (PrimaryKeyColumn, ForeignKeyColumn,
                      StringColumn, IntegerColumn,
                      FloatColumn, BooleanColumn)
from .queries import SelectObject, Select, Insert, Update, Delete


class Model:
    def __init__(self):
        # setting the name that will be used as the table name
        # example str(self) = "<__main__.Test object at 0x000001D9BC34C448>"
        if isinstance(self, Model):
            self.model_name = str(self).split('.')[-1].split(' object ')[0].lower()
            self.get_columns()
        else:
            self.model_name = str(self).split('.')[-1].replace("'>", '').lower()
            self.get_columns(self)

    # get all columns into a dict and assign columns names
    def get_columns(self):
        self.columns = {}
        for var in dir(self):
            # making sure the var is not a magic method
            if '__' not in var:
                class_in = getattr(self, var)
                # checking it is a column
                if isinstance(class_in, (PrimaryKeyColumn, ForeignKeyColumn, IntegerColumn, StringColumn, FloatColumn, BooleanColumn)):
                    self.columns[var] = class_in
                    self.columns[var].name = var
                    if isinstance(class_in, PrimaryKeyColumn):
                        self.primary_key = self.columns[var]

    def check_for_changes(self):
        changes = []

        # checking if the table exists
        table_data = self.session.select(SelectObject("SELECT name FROM sqlite_master WHERE type ='table' AND name = ?;", params=[self.model_name], fetch_mode='one'))
        if table_data:
            # checking that all the columns in the model exist
            """
            table_info keys:
            index : value : type
            0 : cid: int
            1 : name: str
            2 : type: str
            3 : not_null : int/bool
            4 : default : what ever the value of the default was
            5 : primary_key : int/bool
            """
            table_info = self.session.select(SelectObject(f"PRAGMA table_info({self.model_name});", fetch_mode='all'))
            column_name = []
            # checking each column for changes
            for column in self.columns:
                column = self.columns[column]
                column_table_info = None
                for column_ in table_info:
                    if column_[1] == column.name:
                        column_table_info = column_
                if column_table_info:
                    # type check
                    if column.type != column_table_info[2]:
                        changes.append(f'{column.name} is meant to have a type of {column.type} not {column_table_info[2]}')
                    # primary key check
                    if column.pk != bool(column_table_info[5]):
                        if column.pk:
                            changes.append(f'{column.name} is meant to be a primary key')
                        if not column.pk:
                            changes.append(f'{column.name} is not meant to be a primary key')
                    # not null check
                    if column.not_null != bool(column_table_info[3]):
                        if column.not_null:
                            changes.append(f'{column.name} is meant have not_null')
                        else:
                            changes.append(f'{column.name} is not meant to have not_null')
                    # default check
                    if column.default != column_table_info[4]:
                        changes.append(f"{column.name}'s default value is meant to be [{column.default}] not {column_table_info[4]}")
                    # unique check
                    """
                    index_list keys:
                    index : value : type
                    0 : seq : int
                    1 : name : str
                    2 : unique : int/bool
                    3 : origin : str?
                    4 : partial : int/bool
                    """
                    index_list = self.session.select(SelectObject(f'PRAGMA index_list({self.model_name});', fetch_mode='all'))
                    unique = False
                    for index in index_list:
                        """
                        index_info keys:
                        index : value : type
                        0 : seqno : int
                        1 : cid : int
                        2 : name : str
                        """
                        index_info = self.session.select(SelectObject(f'PRAGMA index_info({index[1]});', fetch_mode='one'))
                        if column.name == index_info[2]:
                            unique = True
                    if column.unique and not unique:
                        changes.append(f"{column.name} is meant to be unique")
                    elif not column.unique and unique:
                        changes.append(f"{column.name} is not meant to be unique")
                    # foreign key check
                    """
                    pragma_foreign_key_list keys:
                    index : value : type
                    0 : id : int
                    1 : seq : int
                    2 : table : str
                    3 : from : str
                    4 : to : str
                    5 : on_update : str
                    6 : on_delete : str
                    7 : match : str
                    """
                    foreign_key_list = self.session.select(SelectObject(f"SELECT * FROM pragma_foreign_key_list('{self.model_name}');", fetch_mode='all'))
                    is_foreign_key = False
                    for foreign_key in foreign_key_list:
                        if column.name == foreign_key[3]:
                            is_foreign_key = True
                            foreign_key_info = foreign_key
                    if column.fk and not is_foreign_key:
                        changes.append(f"{column.name} is meant to be foreign key")
                    elif not column.fk and is_foreign_key:
                        changes.append(f"{column.name} is not meant to be foreign key")
                    if is_foreign_key:
                        # foreign table check
                        if column.foreign_model.model_name != foreign_key_info[2]:
                            changes.append(f"{column.name}'s foreign_model is meant to be {column.foreign_model.model_name} but is {foreign_key_info[2]}")
                        # foreign column check
                        if column.foreign_model.primary_key.name != foreign_key_info[4]:
                            changes.append(f"{column.name}'s foreign column is meant to be {column.foreign_model.primary_key.name} but is {foreign_key_info[4]}")
                        # on delete check
                        if column.on_delete.sql != foreign_key_info[6]:
                            changes.append(f"{column.name}'s on_delete is meant to be {column.on_delete.sql} but is {foreign_key_info[6]}")
                        # on update check
                        if column.on_update.sql != foreign_key_info[5]:
                            changes.append(f"{column.name}'s on_update is meant to be {column.on_update.sql} but is {foreign_key_info[5]}")
                        # match check
                        if column.match.sql != foreign_key_info[7]:
                            changes.append(f"{column.name}'s match is meant to be {column.match.sql} but is {foreign_key_info[7]}")
                else:
                    changes.append(f'{column.name} does not exist')
        else:
            changes.append(f"{self.model_name} table does not exist")
        return changes

    def generate_migrations(self, changes):
        table_creation_str = f"CREATE TABLE {self.model_name} ({self.columns[self.primary_key.name].generate_sql()}, "
        for column in self.columns:
            if self.columns[column].pk is False:
                table_creation_str += self.columns[column].generate_sql() + ", "
        table_creation_str = table_creation_str[:-2] + ')'
        if f"{self.model_name} table does not exist" in changes:
            action = 'CREATE'
        else:
            action = 'UPDATE'
        migrations_file_str = f"""\n\n
class {self.model_name}(Migration):
    def __init__(self):
        self.table_name = '{self.model_name}'
        self.table_creation_str = '{table_creation_str}'
        self.changes = {changes}
        self.columns = {list(self.columns.keys())}
        self.action = '{action}'"""
        return migrations_file_str

    def load_queries(self):
        self.select = Select(self.session, self.model_name, self.columns)
        self.insert = Insert(self.session, self.model_name).add
        self.update = Update(self.session, self.model_name).set
        self.delete = Delete(self.session, self.model_name).remove



from .queries import SelectObject, InsertObject


class Migration:
    def create(self):
        self.session.add(InsertObject(self.table_creation_str, commit=True))

    def update(self):
        # getting data before removing the table
        table_info = self.session.select(SelectObject(f"PRAGMA table_info({self.table_name});"), fetch_mode='all')
        table_data = self.session.select(SelectObject(f"SELECT * FROM {self.table_name}"), fetch_mode='all')

        # remaking the table
        self.session.add(InsertObject(f"DROP TABLE {self.table_name}", commit=True))
        self.session.add(InsertObject(self.table_creation_str, commit=True))

        # making the query to add all the data back into the table
        columns_str = ""
        values_str = "("
        for column in table_info:
            columns_str += column[1] + ', '
            values_str += '?, '
        values_str = values_str[:-2] + ')'
        f_value_str = ""
        for row in table_info:
            f_value_str += values_str + ', '

        self.session.add(InsertObject(f"INSERT INTO {self.table_name} ({columns_str[:-2]}) VALUES {f_value_str[:-2]};", params=table_data, commit=True))

    def run_migrations(self, db_path, session):
        self.db_path = db_path
        self.session = session
        if self.action == 'CREATE':
            self.create()
        elif self.action == 'UPDATE':
            self.update()
        else:
            print(f"action '{self.action}' is not recognized.")



from .session import Session
from .version import version
import os
import datetime

dev = False
version_str = f"SQL-Cable Build {version}{'-dev' if dev else ''}"


def get_lastest_migration_file(new=False):
    # gets all files with migration_ in the name
    all_files = [f for f in os.listdir('migrations') if os.path.isfile(os.path.join('migrations', f)) and 'migration_' in f and f != '__init__.py']
    if len(all_files) > 0:
        # sorts the list by number to find the lastest one and make that into a migration file name
        if new:
            file_name = "migration_" + str(int(sorted(all_files, reverse=True)[0].replace('migration_', '').replace('.py', '')) + 1) + ".py"
        else:
            file_name = "migration_" + str(int(sorted(all_files, reverse=True)[0].replace('migration_', '').replace('.py', ''))) + ".py"
    else:
        file_name = "migration_" + '1' + ".py"
    return file_name


class DB:
    def __init__(self, app=None):
        self.models = []
        if app:
            self.init(app)

    def init(self, app):
        self.db_path = app.config['DB_PATH']
        # start the db instances session
        # NOTE there only should ever be one made
        self.session = Session(self.db_path)
        if dev:
            print(f"""WARNING you are running a dev build of SQL-Cable build version: {version}-dev
This build maybe unstable and have bugs or other problems
Please update to the latest full build of SQL-Cable by
typing 'pip install sql-cable' in the cmd""")

    def register_model(self, model):
        # model = model()
        model.db_path = self.db_path
        model.session = self.session
        model.__init__(model)
        model.load_queries(model)
        model.check_for_changes(model)
        self.models.append(model)

    def make_migrations(self, path=None, test=False):
        # if migrations folder does not exist make one and a __init__.py file
        if not os.path.isdir('migrations'):
            os.mkdir('migrations')
            with open(path if path else os.path.join('migrations', '__init__.py'), 'w') as f:
                pass
        # getting the new migration files file name
        file_name = get_lastest_migration_file(new=True)
        file_str = f"# Generated by {version_str} on {str(datetime.datetime.now())[:-7]}\nfrom sql_cable.migrations import Migration"
        # getting all the migrations from models and adding them to the file str
        model_names = []
        for model in self.models:
            changes = model.check_for_changes(model)
            if changes != []:
                migrations_str = model.generate_migrations(model, changes)
                file_str += migrations_str
                model_names.append(f'{model.model_name}()')
        if not model_names:
            print("No migrations to be made")
            return None
        # making the model_names list into a list of instances of the migrations class's
        # eg changing models = models = ['test1()'] to models = [test1()]
        file_str += "\n\n\nmodels = " + str(model_names).replace("'", '')

        # making the migration file and adding it to the __init__.py file
        file_path = os.path.join("migrations", file_name)
        with open(path if path else file_path, 'w') as f:
            f.write(file_str)
        with open(path if path else os.path.join("migrations", "__init__.py"), 'a') as f:
            f.write(f"\nfrom . import {file_name[:-3]}")
        if test:
            return file_path

    def migrate(self):
        if not os.path.isdir('migrations'):
            print("No dir named migrations could be found.")
            return None
        # gettings the latest migration file and imports the migrations package
        migrate_file_name = get_lastest_migration_file()
        migrate_file = __import__("migrations")
        # gets all the models that have changes
        _locals = locals()
        exec(f"models = migrate_file.{migrate_file_name[:-3]}.models", _locals)
        for model in _locals['models']:
            model.run_migrations(self.db_path, self.session)



from .. import DB, Model
from ..columns import (PrimaryKeyColumn, ForeignKeyColumn,
                       StringColumn, IntegerColumn,
                       FloatColumn, BooleanColumn)
from ..queries.query_objects import FilterObject
from .req import get_test_str, get_default_value
import os
import shutil
import sqlite3


class FinalFunctionsTest:
    def __init__(self, app, fetch_mode):
        self.test_name = 'FinalFunctionsTest'
        self.app = app
        self.fetch_mode = fetch_mode

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)
        if self.fetch_mode == 'all':
            result = Test1.select.all()
            self.db.session.stop()
            if len(result) != 8:
                return False
            for i in range(len(result)):
                if result[i].username != 'testuser' + str(i + 1):
                    return False
                if result[i].password != 'password' + str(i + 1):
                    return False
        if self.fetch_mode == 'one':
            result = Test1.select.one()
            self.db.session.stop()
            if result.username != 'testuser1' and result.password != 'password1':
                return False
        return True


class FinalFunctionsEmptyTest:
    def __init__(self, app, fetch_mode):
        self.test_name = 'FinalFunctionsEmptyTest'
        self.app = app
        self.fetch_mode = fetch_mode

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)
        if self.fetch_mode == 'all':
            result = Test1.select.all()
            self.db.session.stop()
            if not result:
                return True
        if self.fetch_mode == 'one':
            result = Test1.select.one()
            self.db.session.stop()
            if result.id is None and result.username is None and result.password is None:
                return True
        return False


class ForeignKeySelectTest:
    def __init__(self, app):
        self.test_name = 'ForeignKeySelectTest'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, name STRING, password STRING);")
        conn.commit()
        c.execute("CREATE TABLE test2 (id INTEGER PRIMARY KEY AUTOINCREMENT, test1_id REFERENCES test1 ([id]) ON DELETE CASCADE);")
        conn.commit()
        c.execute("INSERT INTO test1 (id, name, password) VALUES (?, ?, ?)", (1, 'name1', 'password1'))
        conn.commit()
        c.execute("INSERT INTO test2 (test1_id) VALUES (?)", (1,))
        conn.commit()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            name = StringColumn()
            password = StringColumn()

        class Test2(Model):
            id = PrimaryKeyColumn()
            test1_id = ForeignKeyColumn(Test1)

        self.db.register_model(Test1)
        self.db.register_model(Test2)

        select_data = Test2.select.one()
        self.db.session.stop()
        if select_data.test1_id.rows[0].name != 'name1':
            return False
        if select_data.test1_id.rows[0].password != 'password1':
            return False
        if select_data.test1_id.value != 1:
            return False
        return True


class SortByTests:
    def __init__(self, app):
        self.test_name = 'SortByTests'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()

        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        result = Test1.select.sortby(id='DESC').all()
        self.db.session.stop()
        correct_number = 8
        for i in range(8):
            if result[i].id != correct_number:
                return False
            correct_number -= 1
        return True


class FilterTests:
    def __init__(self, app, fetch_mode):
        self.test_name = 'FilterTests'
        self.app = app
        self.fetch_mode = fetch_mode

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        if self.fetch_mode == 'all':
            result = Test1.select.filter(username='testuser4').all()
            self.db.session.stop()
            if len(result) != 1:
                return False
            if result[0].username != 'testuser4' or result[0].password != 'password4':
                return False
            return True
        elif self.fetch_mode == 'one':
            result = Test1.select.filter(username='testuser4').one()
            self.db.session.stop()
            if result.username != 'testuser4' or result.password != 'password4':
                return False
            return True


class FilterObjectTests:
    def __init__(self, app, fetch_mode):
        self.test_name = 'FilterObjectTests'
        self.app = app
        self.fetch_mode = fetch_mode

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        if self.fetch_mode == 'all':
            result = Test1.select.filter(username1=FilterObject(username='testuser4')).all()
            self.db.session.stop()
            if len(result) != 1:
                return False
            if result[0].username != 'testuser4' or result[0].password != 'password4':
                return False
            return True
        elif self.fetch_mode == 'one':
            result = Test1.select.filter(username1=FilterObject(username='testuser4')).one()
            self.db.session.stop()
            if result.username != 'testuser4' or result.password != 'password4':
                return False
            return True


class MultipoolFiltersTests:
    def __init__(self, app):
        self.test_name = 'MultipoolFiltersTests'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        results = Test1.select.filter(username1=FilterObject(username='testuser4'), username2=FilterObject(username='testuser5'), conjunctive_op='OR').all()
        self.db.session.stop()
        if len(results) != 2:
            return False
        for result in results:
            if result.username != 'testuser4' and result.username != 'testuser5':
                return False
        return True


class FilterComparisonOpTest:
    def __init__(self, app):
        self.test_name = 'FilterComparisonOpTest'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        results = Test1.select.filter(id=4, comparison_op='>').all()
        self.db.session.stop()
        if len(results) != 4:
            return False
        for result in results:
            if result.id <= 4:
                return False
        return True


class FilterSortTest:
    def __init__(self, app):
        self.test_name = 'FilterSortTest'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("CREATE TABLE test1 (id INTEGER PRIMARY KEY AUTOINCREMENT, username STRING, password STRING);")
        conn.commit()
        value_str = ""
        params = []
        for i in range(8):
            value_str += "(?, ?), "
            params.append('testuser' + str(i + 1))
            params.append('password' + str(i + 1))
        c.execute(f"INSERT INTO test1 (username, password) VALUES {value_str[:-2]}", tuple(params))
        conn.commit()
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            username = StringColumn()
            password = StringColumn()

        self.db.register_model(Test1)

        results = Test1.select.filter(username='testuser3', password='password4', conjunctive_op='OR').sortby(id='DESC').all()
        self.db.session.stop()
        correct_number = 4
        if len(results) != 2:
            return False
        for i in range(len(results)):
            if results[i].id != correct_number:
                return False
            if results[i].username != 'testuser3' and results[i].password != 'password4':
                return False
            correct_number -= 1
        return True


class Select:
    def __init__(self, app):
        self.name = 'SelectTests'
        self.app = app
        self.tests = [FinalFunctionsTest(self.app, 'all'),
                      FinalFunctionsTest(self.app, 'one'),
                      FinalFunctionsEmptyTest(self.app, 'all'),
                      FinalFunctionsEmptyTest(self.app, 'one'),
                      ForeignKeySelectTest(self.app),
                      SortByTests(self.app),
                      FilterTests(self.app, 'all'),
                      FilterTests(self.app, 'one'),
                      FilterObjectTests(self.app, 'all'),
                      FilterObjectTests(self.app, 'one'),
                      MultipoolFiltersTests(self.app),
                      FilterComparisonOpTest(self.app),
                      FilterSortTest(self.app)]

    def run(self):
        output = ""
        tests_run = 0
        tests_passed = 0
        for test in self.tests:
            tests_run += 1
            try:
                result = test.run()
                output += get_test_str(test.test_name, result)
                if result:
                    tests_passed += 1
            except Exception as e:
                test.run()
        return output, tests_run, tests_passed



from .. import DB, Model
from ..columns import (PrimaryKeyColumn, ForeignKeyColumn,
                       StringColumn, IntegerColumn,
                       FloatColumn, BooleanColumn)
from .req import get_test_str, get_default_value
import os
import shutil
import sqlite3
import time


two_model_migration = '''
from sql_cable.migrations import Migration


class test1(Migration):
    def __init__(self):
        self.table_name = 'test1'
        self.table_creation_str = 'CREATE TABLE test1 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [name] STRING)'
        self.changes = ['test1 table does not exist']
        self.columns = ['id', 'name']
        self.action = 'CREATE'


class test2(Migration):
    def __init__(self):
        self.table_name = 'test2'
        self.table_creation_str = 'CREATE TABLE test2 ([id] INTEGER PRIMARY KEY AUTOINCREMENT, [test_1_id] INTEGER REFERENCES test1 (id)  ON DELETE CASCADE)'
        self.changes = ['test2 table does not exist']
        self.columns = ['id', 'test_1_id']
        self.action = 'CREATE'


models = [test1(), test2()]'''


class TwoModelMigration:
    def __init__(self, app):
        self.test_name = 'TwoModelMigration'
        self.app = app

    def run(self):
        if os.path.isdir('migrations'):
            shutil.rmtree('migrations')
        if os.path.isfile(self.app.config['DB_PATH']):
            os.remove(self.app.config['DB_PATH'])
        self.db = DB(self.app)

        class Test1(Model):
            id = PrimaryKeyColumn()
            name = StringColumn()

        class Test2(Model):
            id = PrimaryKeyColumn()
            test_1_id = ForeignKeyColumn(Test1)

        self.db.register_model(Test1)
        self.db.register_model(Test2)

        if not os.path.isdir('migrations'):
            os.mkdir('migrations')
            with open(os.path.join('migrations', '__init__.py'), 'w') as f:
                f.write("\nfrom . import migration_1")
        with open(os.path.join('migrations', 'migration_1.py'), 'w') as f:
            f.write(two_model_migration)
        self.db.migrate()
        self.db.session.stop()

        conn = sqlite3.connect(self.app.config['DB_PATH'])
        c = conn.cursor()
        c.execute("PRAGMA table_info(test1);")
        table_info = c.fetchall()
        if table_info[0][1] == 'id' and table_info[1][1] == 'name':
            c.execute("PRAGMA table_info(test2);")
            table_info = c.fetchall()
            conn.close()
            if table_info[0][1] == 'id' and table_info[1][1] == 'test_1_id':
                return True
        return False


class MigrateTests:
    def __init__(self, app):
        self.name = 'MigrateTests'
        self.app = app
        self.tests = [TwoModelMigration(self.app)]

    def run(self):
        output = ""
        tests_run = 0
        tests_passed = 0
        for test in self.tests:
            tests_run += 1
            try:
                result = test.run()
                output += get_test_str(test.test_name, result)
                if result:
                    tests_passed += 1
            except Exception as e:
                test.run()
        return output, tests_run, tests_passed
